%LINE_SE   Morphological operations with a line segment as SE.
%
%   OUT       = LINE_SE(OPERATION,IN,L,PHI,METHOD)
%   OPERATION = 'dilation', 'closing',  etc.
%   L,PHI     = length and angle of line to be used as SE
%   METHOD    = method to use:
%                 'discrete':          direct implementation
%                 'recursive':         along bresenhman lines
%                 'soille':            intersection of closings / union of openings
%                 'periodic':          with periodic line
%                 'interpolated':      along interpolated lines
%                 'true_interpolated': with interpolation
%                 'grey-value':        grey-value SE

function out = line_se(operation,in,l,phi,method)
if nargin<5
   method = 'bresenham_line';
   if nargin<4
      phi = 0;
      if nargin<3
         l = 5;
         if nargin<2
            error('input image required.')
         end
      end
   end
end
operation = lower(operation);
if ~any(strcmp({'dilation','erosion','closing','opening'},operation))
   error('Unknown operation.')
end
if length(l) ~= 1 | length(phi) ~= 1
   error('L and PHI parameters must be scalars')
end
bin = islogical(in);
if bin
   in = +in;
end
switch lower(method)
   case {'discrete','discrete_line'}
      l = round(l*[cos(phi),sin(phi)]);
      start = [0,0];           % start point of line
      stop = l-sign(l);        % end point of line
      sz = abs(stop)+1;
      sz = sz + ~mod(sz,2);    % image is odd in size.
      se = newim(sz,'bin');
      ii = stop<0;
      start(ii) = -stop(ii);   % change start/stop so that line is in image
      stop(ii) = 0;
      se = drawline(se,start,stop,1);
      switch operation
         case 'dilation'
            out = dip_dilation(in,se,[0,0],'user_defined');
         case 'erosion'
            out = dip_erosion(in,se,[0,0],'user_defined');
         case 'closing'
            out = dip_closing(in,se,[0,0],'user_defined');
         case 'opening'
            out = dip_opening(in,se,[0,0],'user_defined');
      end
   case {'bresenham','bresenham_line','recursive','vanherk'}
      switch operation
         case 'dilation'
            out = dip_dilation(in,[],[l,phi],'discrete_line');
         case 'erosion'
            out = dip_erosion(in,[],[l,phi],'discrete_line');
         case 'closing'
            out = dip_closing(in,[],[l,phi],'discrete_line');
         case 'opening'
            out = dip_opening(in,[],[l,phi],'discrete_line');
      end
   case 'soille'
      if any(strcmp({'dilation','erosion'},operation))
         error('Dilation and erosion not possible with method ''soille''.')
      end
      out = soille_line(in,l,phi,operation,1);
   case 'soille2'
      if any(strcmp({'dilation','erosion'},operation))
         error('Dilation and erosion not possible with method ''soille2''.')
      end
      out = soille_line(in,l,phi,operation,2);
   case {'periodic','periodic_line'}
      % We don't have the quick implementation, so this is really useless. But what the heck!
      l = round(l*[cos(phi),sin(phi)]);
      l = l-sign(l);
      n = gcd(l(1),l(2));
      p = l/n;
      start = [0,0];           % start point of line
      stop = l;                % end point of line
      sz = abs(stop)+1;
      sz = sz + ~mod(sz,2);    % image is odd in size.
      se = newim(sz,'bin');
      ii = stop<0;
      start(ii) = -stop(ii);   % change start/stop so that line is in image
      stop(ii) = 0;
      for ii=0:n
         se(start(1)+p(1)*ii,start(2)+p(2)*ii) = 1;
      end
      switch operation
         case 'dilation'
            out = dip_dilation(in,se,[0,0],'user_defined');
         case 'erosion'
            out = dip_erosion(in,se,[0,0],'user_defined');
         case 'closing'
            out = dip_closing(in,se,[0,0],'user_defined');
         case 'opening'
            out = dip_opening(in,se,[0,0],'user_defined');
      end
   case {'interpolated','interpolated_line'}
      switch operation
         case 'dilation'
            out = dip_dilation(in,[],[l,phi],'interpolated_line');
         case 'erosion'
            out = dip_erosion(in,[],[l,phi],'interpolated_line');
         case 'closing'
            out = dip_closing(in,[],[l,phi],'interpolated_line');
         case 'opening'
            out = dip_opening(in,[],[l,phi],'interpolated_line');
      end
   case {'true_interpolated','true_interpolated_line'}
      out = interpolatedline(operation,in,phi,l);
   case {'grey','grey-value','gray','gray-value'}
      sz = ceil(abs(l*[cos(phi),sin(phi)]))+4;
      sz = sz + ~mod(sz,2);    % image is odd in size.
      se = bandlimitedline(sz,l,phi);
      se = (1.0853*(max(in)-min(in)))*se;
      in = dip_image(in,'sfloat');
      switch operation
         case 'dilation'
            out = dip_dilation(in,se,[0,0],'user_defined');
         case 'erosion'
            out = dip_erosion(in,se,[0,0],'user_defined');
         case 'closing'
            out = dip_closing(in,se,[0,0],'user_defined');
         case 'opening'
            out = dip_opening(in,se,[0,0],'user_defined');
      end
   otherwise
      error('Unknown method.')
end
if bin
   out = dip_image(out,'bin');
end


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function out = bandlimitedline(in,l,phi)
x = xx(in);
y = yy(in);
rx = x*cos(phi) + y*sin(phi);
rx = (l/2)-abs(rx);
ry = y*cos(phi) - x*sin(phi);
rx(abs(ry)>4) = -4; % avoid computations
ry(rx<-4) = -4;     % avoid computations
a1 = gaussianedgeclip(rx,1);
a2 = gaussianlineclip(ry,1,1);
out = a1*a2 - 1;


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function out = interpolatedline(operation,in,phi,l)
sz = size(in);
phi = mod(phi+pi/4,pi) - pi/4; % a different range for phi than above
if phi>pi/4  % phi in (pi/4,3pi/4)
   skew = 0;
   axis = 1;
   phi = pi/2 - phi;
   l = l*cos(phi);
else         % phi in (-pi/4,pi/4)
   skew = 1;
   axis = 0;
   %phi = -phi;
   l = l*cos(phi);
end
l = round(l);
switch operation
   case {'dilation','erosion'}
      s = ceil((l-1)/2);
   case {'closing','opening'}
      s = l-1;
end
out = newim(in);
if axis==1 % Horizontal skews, vertical operation
   for ii=0:sz(2)-1
      first = max(ii-s,0);
      last = min(ii+s,sz(2)-1);
      buf = newim(sz(1),2*s+1);
      buf(:,first-(ii-s):last-(ii-s)) = in(:,first:last);
      buf = dip_skewing(buf,phi,skew,axis,'3-cubic','zero',0);
      switch operation
         case 'dilation'
            buf = dip_dilation(buf,[],[1,l],'rectangular');
         case 'erosion'
            buf = dip_erosion(buf,[],[1,l],'rectangular');
         case 'closing'
            buf = dip_closing(buf,[],[1,l],'rectangular');
         case 'opening'
            buf = dip_opening(buf,[],[1,l],'rectangular');
      end
      skip = ceil((size(buf,1)-sz(1))/2);
      out(:,ii) = buf(skip+[0:sz(1)-1],(size(buf,2)-1)/2);
   end
else       % Vertical skews, horizontal operation
   for ii=0:sz(1)-1
      first = max(ii-s,0);
      last = min(ii+s,sz(1)-1);
      buf = newim(2*s+1,sz(2));
      buf(first-(ii-s):last-(ii-s),:) = in(first:last,:);
      buf = dip_skewing(buf,phi,skew,axis,'3-cubic','zero',0);
      switch operation
         case 'dilation'
            buf = dip_dilation(buf,[],[l,1],'rectangular');
         case 'erosion'
            buf = dip_erosion(buf,[],[l,1],'rectangular');
         case 'closing'
            buf = dip_closing(buf,[],[l,1],'rectangular');
         case 'opening'
            buf = dip_opening(buf,[],[l,1],'rectangular');
      end
      skip = ceil((size(buf,2)-sz(2))/2);
      out(ii,:) = buf((size(buf,1)-1)/2,skip+[0:sz(2)-1]);
   end
end


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function out = soille_line(in,l,phi,operation,version)
switch operation
   case 'closing'
      filter = 'dip_closing';
      operation = 'min';
   case 'opening'
      filter = 'dip_opening';
      operation = 'max';
   otherwise
      error('Operation not implemented')
end
phi = mod(phi+pi/4,pi) - pi/4;  % forcing angle to be in [-pi/4,3pi/4] range
if phi==0
   N = 1;
else
   if version == 1
      rect = round(l*[cos(phi),sin(phi)]);
      l = norm(rect);                 % rounding of length
      phi = atan2(rect(2),rect(1));   % rounding of angle
      phi = mod(phi+pi/4,pi) - pi/4;  % forcing angle to be in [-pi/4,3pi/4] range
   else
      sz = size(in);
      rect = round(min(sz(1)*[1,tan(phi)],sz(2)*[1/tan(phi),1]));
   end
   N = periodicity(abs(rect));
end
if phi > pi/4
   % Vertical wrapping
   wparam = [0,1];
   cropop = 'tmp(:,ii-1:end-N+ii-1)';
   in = [in;newim([size(in,1),N],datatype(in))];   
else
   % Horizontal wrapping
   wparam = [1,0];
   cropop = 'tmp(ii-1:end-N+ii-1,:)';
   in = [in,newim([N,size(in,2)],datatype(in))];
end
ii = 1;
tmp = feval(filter,in,[],[l,phi],'discrete_line');
out = eval(cropop);
for ii=2:N
   in = dip_wrap(in,wparam);
   tmp = feval(filter,in,[],[l,phi],'discrete_line');
   tmp = eval(cropop);
   out = feval(operation,out,tmp);
end


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function p = periodicity(rect)
if any(mod(rect,1)~=0)
   error('rect must be integer valued')
end
q = max(rect);
r = min(rect);
if r==0
   p = 1;
else
   p = q;
   for d = r:-1:2
      if mod(r/d,1)==0 & mod(q/d,1)==0
         p = q/d;
         break;
      end
   end
end
